;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "forth-mode" "20260315.803"
  "Programming language mode for Forth."
  '((cl-lib "0.2"))
  :url "https://github.com/larsbrinkhoff/forth-mode"
  :commit "4450a3a5629b579f5d2045d0d8aec84193e9a31f"
  :revdesc "4450a3a5629b"
  :keywords '("languages" "forth")
  :authors '(("Lars Brinkhoff" . "lars@nocrew.org"))
  :maintainers '(("Lars Brinkhoff" . "lars@nocrew.org")))
